/*
 * Name:    d_mcsscl.c
 *
 * Purpose: MX scaler driver to control MCS channels as individual scalers.
 *
 *          By convention, the scaler number for an MCS scaler should match
 *          the labeling of scaler channel numbers on the front panel (if any)
 *          of the MCS.  It is the responsibility of the underlying MCS
 *          record to put the data at an appropriate place in mcs->data_array.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_meas.h"
#include "mx_scalr.h"
#include "mx_mcs.h"
#include "d_mcsscl.h"

/* Initialize the scaler driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_mcs_scaler_record_function_list = {
	mxd_mcs_scaler_initialize_type,
	mxd_mcs_scaler_create_record_structures,
	mxd_mcs_scaler_finish_record_initialization,
	mxd_mcs_scaler_delete_record,
	mxd_mcs_scaler_print_structure,
	mxd_mcs_scaler_read_parms_from_hardware,
	mxd_mcs_scaler_write_parms_to_hardware,
	mxd_mcs_scaler_open,
	mxd_mcs_scaler_close
};

MX_SCALER_FUNCTION_LIST mxd_mcs_scaler_scaler_function_list = {
	mxd_mcs_scaler_clear,
	mxd_mcs_scaler_overflow_set,
	mxd_mcs_scaler_read,
	mxd_mcs_scaler_read_and_clear,
	mxd_mcs_scaler_is_busy,
	mxd_mcs_scaler_start,
	mxd_mcs_scaler_stop,
	mxd_mcs_scaler_get_mode,
	mxd_mcs_scaler_set_mode,
	mxd_mcs_scaler_set_modes_of_associated_counters
};

/* MCS scaler data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_mcs_scaler_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_SCALER_STANDARD_FIELDS,
	MXD_MCS_SCALER_STANDARD_FIELDS
};

long mxd_mcs_scaler_num_record_fields
		= sizeof( mxd_mcs_scaler_record_field_defaults )
		  / sizeof( mxd_mcs_scaler_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_mcs_scaler_rfield_def_ptr
			= &mxd_mcs_scaler_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_mcs_scaler_get_pointers( MX_SCALER *scaler,
			MX_MCS_SCALER **mcs_scaler,
			const char *calling_fname )
{
	const char fname[] = "mxd_mcs_scaler_get_pointers()";

	if ( scaler == (MX_SCALER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_SCALER pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( mcs_scaler == (MX_MCS_SCALER **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_MCS_SCALER pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*mcs_scaler = (MX_MCS_SCALER *) (scaler->record->record_type_struct);

	if ( *mcs_scaler == (MX_MCS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCS_SCALER pointer for record '%s' is NULL.",
			scaler->record->name );
	}

	return MX_SUCCESSFUL_RESULT;
}

/*------------------------------------------------------------------*/

MX_EXPORT mx_status_type
mxd_mcs_scaler_initialize_type( long type )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_mcs_scaler_create_record_structures()";

	MX_SCALER *scaler;
	MX_MCS_SCALER *mcs_scaler;

	/* Allocate memory for the necessary structures. */

	scaler = (MX_SCALER *) malloc( sizeof(MX_SCALER) );

	if ( scaler == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SCALER structure." );
	}

	mcs_scaler = (MX_MCS_SCALER *)
				malloc( sizeof(MX_MCS_SCALER) );

	if ( mcs_scaler == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MCS_SCALER structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = scaler;
	record->record_type_struct = mcs_scaler;
	record->class_specific_function_list
			= &mxd_mcs_scaler_scaler_function_list;

	scaler->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxd_mcs_scaler_finish_record_initialization()";

	MX_RECORD *mcs_record;
	MX_MCS *mcs;
	MX_SCALER *scaler;
	MX_MCS_SCALER *mcs_scaler;
	long scaler_index;

	scaler = (MX_SCALER *) record->record_class_struct;

	if ( scaler == (MX_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_SCALER pointer for record '%s' is NULL.",
			record->name );
	}

	scaler->automatic_overflow_check = FALSE;

	mcs_scaler = (MX_MCS_SCALER *) record->record_type_struct;

	if ( mcs_scaler == (MX_MCS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCS_SCALER pointer for record '%s' is NULL.",
			record->name );
	}

	mcs_record = mcs_scaler->mcs_record;

	if ( mcs_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The mcs_record pointer for MCS scaler '%s' is NULL.",
			record->name );
	}

	mcs = (MX_MCS *) mcs_record->record_class_struct;

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCS pointer for MCS record '%s' is NULL.",
			mcs_record->name );
	}

	if ( mcs->scaler_record_array == (MX_RECORD **) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The scaler_record_array pointer for MCS record '%s' is NULL.",
			mcs_record->name );
	}

	scaler_index = mcs_scaler->scaler_number - mcs->first_scaler_number;

	mcs->scaler_record_array[ scaler_index ] = record;

#if 0
	MX_DEBUG(-2,("%s: record = %p", fname, record));
	MX_DEBUG(-2,("%s: record->name = '%s'", fname, record->name));
	MX_DEBUG(-2,("%s: mcs_scaler = %p", fname, mcs_scaler));
	MX_DEBUG(-2,("%s: mcs_scaler->mcs_record = %p",
					fname, mcs_scaler->mcs_record));
	MX_DEBUG(-2,("%s: mcs_scaler->mcs_record->name = '%s'",
					fname, mcs_scaler->mcs_record->name));
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_mcs_scaler_print_structure()";

	MX_SCALER *scaler;
	MX_MCS_SCALER *mcs_scaler;
	long current_value;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed is NULL.");
	}

	scaler = (MX_SCALER *) (record->record_class_struct);

	if ( scaler == (MX_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_SCALER pointer for record '%s' is NULL.", record->name);
	}

	mcs_scaler = (MX_MCS_SCALER *) (record->record_type_struct);

	if ( mcs_scaler == (MX_MCS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_MCS_SCALER pointer for record '%s' is NULL.", record->name);
	}

	fprintf(file, "SCALER parameters for scaler '%s':\n", record->name);

	fprintf(file, "  Scaler type           = MCS_SCALER.\n\n");
	fprintf(file, "  MCS record            = %s\n",
					mcs_scaler->mcs_record->name);
	fprintf(file, "  scaler number         = %d\n",
					mcs_scaler->scaler_number);

	status = mx_scaler_read( record, &current_value );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read value of scaler '%s'",
			record->name );
	}

	fprintf(file, "  present value         = %ld\n", current_value);
	fprintf(file, "  dark current          = %g counts per second.\n",
						scaler->dark_current);
	fprintf(file, "  subtract_dark_current = %d\n",
						scaler->subtract_dark_current);

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_open( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_close( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_clear( MX_SCALER *scaler )
{
	const char fname[] = "mxd_mcs_scaler_clear()";

	MX_MCS_SCALER *mcs_scaler;
	mx_status_type mx_status;

	mx_status = mxd_mcs_scaler_get_pointers( scaler, &mcs_scaler, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for scaler '%s'",
			fname, scaler->record->name));

	mx_status = mx_mcs_clear( mcs_scaler->mcs_record );

	MX_DEBUG( 2,("%s complete.", fname));

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_overflow_set( MX_SCALER *scaler )
{
	const char fname[] = "mxd_mcs_scaler_overflow_set()";

	scaler->overflow_set = 0;

	return mx_error( MXE_UNSUPPORTED, fname,
		"This type of scaler cannot check for overflow set." );
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_read( MX_SCALER *scaler )
{
	const char fname[] = "mxd_mcs_scaler_read()";

	MX_MCS_SCALER *mcs_scaler;
	MX_MCS *mcs;
	long scaler_index;
	mx_status_type mx_status;

	mx_status = mxd_mcs_scaler_get_pointers( scaler, &mcs_scaler, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for scaler '%s'",
			fname, scaler->record->name));

	mcs = (MX_MCS *) mcs_scaler->mcs_record->record_class_struct;

	if ( mcs == (MX_MCS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MCS pointer for MCS record '%s' is NULL.",
			mcs_scaler->mcs_record->name );
	}
	if ( mcs->data_array == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"data_array pointer for MCS record '%s' is NULL.",
			mcs_scaler->mcs_record->name );
	}

	mx_status = mx_mcs_set_num_measurements( mcs_scaler->mcs_record, 1 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	scaler_index = mcs_scaler->scaler_number - mcs->first_scaler_number;

	mx_status = mx_mcs_read_scaler( mcs_scaler->mcs_record,
						scaler_index, NULL, NULL );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

#if 0
	MX_DEBUG(-2,("%s: mcs = %p", fname, mcs));
	MX_DEBUG(-2,("%s: mcs->data_array = %p", fname, mcs->data_array));
#endif

	scaler->value = mcs->data_array[ scaler_index ][0];

	MX_DEBUG( 2,("%s complete.  value = %ld", fname, scaler->value));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_read_and_clear( MX_SCALER *scaler )
{
	mx_status_type mx_status;

	mx_status = mxd_mcs_scaler_read( scaler );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxd_mcs_scaler_clear( scaler );

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_is_busy( MX_SCALER *scaler )
{
	const char fname[] = "mxd_mcs_scaler_is_busy()";

	MX_MCS_SCALER *mcs_scaler;
	int busy;
	mx_status_type mx_status;

	mx_status = mxd_mcs_scaler_get_pointers( scaler, &mcs_scaler, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,
	("%s invoked for scaler '%s'", fname, scaler->record->name));

	mx_status = mx_mcs_is_busy( mcs_scaler->mcs_record, &busy );

	scaler->busy = busy;

	MX_DEBUG( 2,("%s complete.  busy = %d", fname, scaler->busy));

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_start( MX_SCALER *scaler )
{
	const char fname[] = "mxd_mcs_scaler_start()";

	MX_MCS_SCALER *mcs_scaler;
	mx_status_type mx_status;

	mx_status = mxd_mcs_scaler_get_pointers( scaler, &mcs_scaler, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s invoked for scaler '%s' for %ld counts.",
			fname, scaler->record->name, scaler->value));

	mx_status = mx_mcs_set_measurement_counts( mcs_scaler->mcs_record,
							scaler->value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s: measurement counts have been set.", fname));

	mx_status = mx_mcs_set_num_measurements( mcs_scaler->mcs_record, 1L );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,("%s: num measurements has been set.", fname));

	mx_status = mx_mcs_start( mcs_scaler->mcs_record );

	MX_DEBUG( 2,("%s complete.", fname));

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_stop( MX_SCALER *scaler )
{
	const char fname[] = "mxd_mcs_scaler_stop()";

	MX_MCS_SCALER *mcs_scaler;
	mx_status_type mx_status;

	mx_status = mxd_mcs_scaler_get_pointers( scaler, &mcs_scaler, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG( 2,
	("%s invoked for scaler '%s'", fname, scaler->record->name));

	mx_status = mx_mcs_stop( mcs_scaler->mcs_record );

	MX_DEBUG( 2,("%s complete.", fname));

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_get_mode( MX_SCALER *scaler )
{
	const char fname[] = "mxd_mcs_scaler_get_mode()";

	MX_MCS_SCALER *mcs_scaler;
	int mode;
	mx_status_type mx_status;

	mx_status = mxd_mcs_scaler_get_pointers( scaler, &mcs_scaler, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mx_mcs_get_mode( mcs_scaler->mcs_record, &mode );

	scaler->mode = mode;

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_set_mode( MX_SCALER *scaler )
{
	const char fname[] = "mxd_mcs_scaler_set_mode()";

	MX_MCS_SCALER *mcs_scaler;
	mx_status_type mx_status;

	mx_status = mxd_mcs_scaler_get_pointers( scaler, &mcs_scaler, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* The following is currently disabled since it is not clear that
	 * running an MCS in a non-preset time mode makes any sense.
	 */
#if 0
	mx_status = mx_mcs_set_mode( mcs_scaler->mcs_record, scaler->mode );
#else
	mx_status = MX_SUCCESSFUL_RESULT;
#endif

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_mcs_scaler_set_modes_of_associated_counters( MX_SCALER *scaler )
{
	/* This function is unnecessary if all the counters for a measurement
	 * belong to the same multichannel scaler (which is true at present).
	 */

	return MX_SUCCESSFUL_RESULT;
}

