/*
 * Name:    mx_epics.c
 *
 * Purpose: Some functions for using EPICS with MX.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999-2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"

#if HAVE_EPICS

/* The following three headers from EPICS. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

#include "mx_rec.h"
#include "mx_epics.h"

/* The variable mx_ezca_is_initialized is used to determine whether or not
 * we need to invoke ezcaAutoErrorMessageOff().
 */

static int mx_ezca_is_initialized = FALSE;

/* The variable mx_epics_debug_flag is used to control whether or not
 * EPICS debugging messages are generated.
 */

static int mx_epics_debug_flag = FALSE;

MX_EXPORT int
mx_epics_get_debug_flag( void )
{
	if ( mx_epics_debug_flag == FALSE ) {
		return FALSE;
	} else {
		return TRUE;
	}
}

MX_EXPORT void
mx_epics_set_debug_flag( int flag )
{
	if ( flag == FALSE ) {
		mx_epics_debug_flag = FALSE;
	} else {
		mx_epics_debug_flag = TRUE;
	}
}

MX_EXPORT mx_status_type
mx_caget( char *pvname, char epics_type, int num_elements, void *data_buffer )
{
	const char fname[] = "mx_caget()";

	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status;

	if ( mx_ezca_is_initialized == FALSE ) {
		ezcaAutoErrorMessageOff();

		mx_ezca_is_initialized = TRUE;
	}

	if ( mx_epics_debug_flag ) {
		MX_DEBUG(-2,("%s: about to read from '%s'.", fname, pvname));
	}

	ezca_status = ezcaGet( pvname, epics_type, num_elements, data_buffer );

	if ( ezca_status != EZCA_OK ) {

		ezcaGetErrorString( NULL, &ezca_error_message );

		sprintf( mx_error_message,
	"Attempt to read from EPICS PV '%s' failed.  EZCA status = %d (%s )",
			pvname, ezca_status, ezca_error_message );

		ezcaFree( (void *) ezca_error_message );

		return mx_error( MXE_DEVICE_IO_ERROR, fname,
				mx_error_message );
	}

	if ( mx_epics_debug_flag ) {
		if ( num_elements > 1 ) {
			if ( epics_type == ezcaString ) {
				MX_DEBUG(-2,("%s: '%s' value read = '%s'.",
					fname, pvname, (char *) data_buffer));
			} else {
				MX_DEBUG(-2,
				("%s: multielement array read from '%s'.",
					fname, pvname));
			}
		} else {
			char byte_value;
			short short_value;
			long long_value;
			double double_value;

			switch( epics_type ) {
			case ezcaByte:
				byte_value = *((char *) data_buffer);

				MX_DEBUG(-2,("%s: '%s' value read = %d.",
					fname, pvname, byte_value));
				break;
			case ezcaString:
				MX_DEBUG(-2,("%s: '%s' value read = '%s'",
					fname, pvname, (char *) data_buffer));
				break;
			case ezcaShort:
				short_value = *((short *) data_buffer);
	
				MX_DEBUG(-2,("%s: '%s' value read = %hd.",
					fname, pvname, short_value));
				break;
			case ezcaLong:
				long_value = *((long *) data_buffer);
	
				MX_DEBUG(-2,("%s: '%s' value read = %ld.",
					fname, pvname, long_value));
				break;
			case ezcaDouble:
				double_value = *((double *) data_buffer);

				MX_DEBUG(-2,("%s: '%s' value read = %g.",
					fname, pvname, double_value));
				break;
			default:
				MX_DEBUG(-2,
				("%s: '%s' has unsupported epics_type %d",
					fname, pvname, epics_type));
				break;
			}
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_caput( char *pvname, char epics_type, int num_elements, void *data_buffer )
{
	const char fname[] = "mx_caput()";

	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status;

	if ( mx_ezca_is_initialized == FALSE ) {
		ezcaAutoErrorMessageOff();

		mx_ezca_is_initialized = TRUE;
	}

	if ( mx_epics_debug_flag ) {
		if ( num_elements > 1 ) {
			if ( epics_type == ezcaString ) {
				MX_DEBUG(-2,("%s: sending '%s' to '%s'.",
					fname, (char *) data_buffer, pvname));
			} else {
				MX_DEBUG(-2,
				("%s: sending multielement array to '%s'.",
					fname, pvname));
			}
		} else {
			char byte_value;
			short short_value;
			long long_value;
			double double_value;
	
			switch( epics_type ) {
			case ezcaByte:
				byte_value = *((char *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %d to '%s'.",
					fname, byte_value, pvname));
				break;
			case ezcaString:
				MX_DEBUG(-2,("%s: sending '%s' to '%s'",
					fname, (char *) data_buffer, pvname));
				break;
			case ezcaShort:
				short_value = *((short *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %hd to '%s'.",
					fname, short_value, pvname));
				break;
			case ezcaLong:
				long_value = *((long *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %ld to '%s'.",
					fname, long_value, pvname));
				break;
			case ezcaDouble:
				double_value = *((double *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %g to '%s'.",
					fname, double_value, pvname));
				break;
			default:
				MX_DEBUG(-2,
				("%s: '%s' has unsupported epics_type %d",
					fname, pvname, epics_type));
				break;
			}
		}
	}

	ezca_status = ezcaPut( pvname, epics_type, num_elements, data_buffer );

	if ( ezca_status != EZCA_OK ) {

		ezcaGetErrorString( NULL, &ezca_error_message );

		sprintf( mx_error_message,
	"Attempt to write to EPICS PV '%s' failed.  EZCA status = %d (%s )",
			pvname, ezca_status, ezca_error_message );

		ezcaFree( (void *) ezca_error_message );

		return mx_error( MXE_DEVICE_IO_ERROR, fname,
				mx_error_message );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_caput_nowait( char *pvname,
			char epics_type, int num_elements, void *data_buffer )
{
	const char fname[] = "mx_caput_nowait()";

	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status;

	if ( mx_ezca_is_initialized == FALSE ) {
		ezcaAutoErrorMessageOff();

		mx_ezca_is_initialized = TRUE;
	}

	if ( mx_epics_debug_flag ) {
		if ( num_elements > 1 ) {
			if ( epics_type == ezcaString ) {
				MX_DEBUG(-2,("%s: sending '%s' to '%s'.",
					fname, (char *) data_buffer, pvname));
			} else {
				MX_DEBUG(-2,
				("%s: sending multielement array to '%s'.",
					fname, pvname));
			}
		} else {
			char byte_value;
			short short_value;
			long long_value;
			double double_value;
	
			switch( epics_type ) {
			case ezcaByte:
				byte_value = *((char *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %d to '%s'.",
					fname, byte_value, pvname));
				break;
			case ezcaString:
				MX_DEBUG(-2,("%s: sending '%s' to '%s'",
					fname, (char *) data_buffer, pvname));
				break;
			case ezcaShort:
				short_value = *((short *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %hd to '%s'.",
					fname, short_value, pvname));
				break;
			case ezcaLong:
				long_value = *((long *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %ld to '%s'.",
					fname, long_value, pvname));
				break;
			case ezcaDouble:
				double_value = *((double *) data_buffer);
	
				MX_DEBUG(-2,("%s: sending %g to '%s'.",
					fname, double_value, pvname));
				break;
			default:
				MX_DEBUG(-2,
				("%s: '%s' has unsupported epics_type %d",
					fname, pvname, epics_type));
				break;
			}
		}
	}
	
	ezca_status = ezcaPutOldCa( pvname,
				epics_type, num_elements, data_buffer );

	if ( ezca_status != EZCA_OK ) {

		ezcaGetErrorString( NULL, &ezca_error_message );

		sprintf( mx_error_message,
	"Attempt to write to EPICS PV '%s' failed.  EZCA status = %d (%s )",
			pvname, ezca_status, ezca_error_message );

		ezcaFree( (void *) ezca_error_message );

		return mx_error( MXE_DEVICE_IO_ERROR, fname,
				mx_error_message );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_epics_convert_mx_type_to_epics_type( long mx_type, chtype *epics_type )
{
	const char fname[] = "mx_epics_convert_mx_type_to_epics_type()";

	switch( mx_type ) {
	case MXFT_STRING:
		*epics_type = DBR_STRING;
		break;
	case MXFT_CHAR:
	case MXFT_UCHAR:
		*epics_type = DBR_CHAR;
		break;
	case MXFT_SHORT:
	case MXFT_USHORT:
		*epics_type = DBR_SHORT;
		break;
	case MXFT_INT:
	case MXFT_UINT:
	case MXFT_LONG:
	case MXFT_ULONG:
		*epics_type = DBR_LONG;
		break;
	case MXFT_FLOAT:
		*epics_type = DBR_FLOAT;
		break;
	case MXFT_DOUBLE:
		*epics_type = DBR_DOUBLE;
		break;
	default:
		*epics_type = -1;
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"MX record field type of %ld cannot be converted to an EPICS type.",
			mx_type );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_epics_convert_epics_type_to_mx_type( chtype epics_type, long *mx_type )
{
	const char fname[] = "mx_epics_convert_epics_type_to_mx_type()";

	switch( epics_type ) {
	case DBR_STRING:
		*mx_type = MXFT_STRING;
		break;
	case DBR_CHAR:
		*mx_type = MXFT_CHAR;
		break;
	case DBR_SHORT:
		*mx_type = MXFT_SHORT;
		break;
	case DBR_LONG:
		*mx_type = MXFT_LONG;
		break;
	case DBR_FLOAT:
		*mx_type = MXFT_FLOAT;
		break;
	case DBR_DOUBLE:
		*mx_type = MXFT_DOUBLE;
		break;
	default:
		*mx_type = -1;
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"EPICS type of %ld cannot be converted to an MX record field type.",
			mx_type );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_epics_convert_mx_type_to_ezca_type( long mx_type, char *ezca_type )
{
	const char fname[] = "mx_epics_convert_mx_type_to_ezca_type()";

	switch( mx_type ) {
	case MXFT_STRING:
		*ezca_type = ezcaString;
		break;
	case MXFT_CHAR:
	case MXFT_UCHAR:
		*ezca_type = ezcaByte;
		break;
	case MXFT_SHORT:
	case MXFT_USHORT:
		*ezca_type = ezcaShort;
		break;
	case MXFT_INT:
	case MXFT_UINT:
	case MXFT_LONG:
	case MXFT_ULONG:
		*ezca_type = ezcaLong;
		break;
	case MXFT_FLOAT:
		*ezca_type = ezcaFloat;
		break;
	case MXFT_DOUBLE:
		*ezca_type = ezcaDouble;
		break;
	default:
		*ezca_type = -1;
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"MX record field type of %ld cannot be converted to an EZCA type.",
			mx_type );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_epics_convert_ezca_type_to_mx_type( char ezca_type, long *mx_type )
{
	const char fname[] = "mx_epics_convert_ezca_type_to_mx_type()";

	switch( ezca_type ) {
	case ezcaString:
		*mx_type = MXFT_STRING;
		break;
	case ezcaByte:
		*mx_type = MXFT_CHAR;
		break;
	case ezcaShort:
		*mx_type = MXFT_SHORT;
		break;
	case ezcaLong:
		*mx_type = MXFT_LONG;
		break;
	case ezcaFloat:
		*mx_type = MXFT_FLOAT;
		break;
	case ezcaDouble:
		*mx_type = MXFT_DOUBLE;
		break;
	default:
		*mx_type = -1;
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"EZCA type of %d cannot be converted to an MX record field type.",
			ezca_type );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */
