/*
 * Name:    mx_rs232.c
 *
 * Purpose: MX function library of generic RS-232 operations.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999-2000 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "mx_util.h"
#include "mx_rs232.h"
#include "mx_drivr.h"

MX_EXPORT mx_status_type
mx_rs232_check_port_parameters( MX_RECORD *rs232_record )
{
	const char fname[] = "mx_rs232_check_port_parameters()";

	MX_RS232 *rs232;
	int i, j, shift;
	unsigned char terminator_value;
	mx_status_type mx_status;

	mx_status = mx_rs232_get_pointers( rs232_record, &rs232, NULL, fname );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Check for allowed serial port speeds. */

	switch( rs232->speed ) {
	case 300:
	case 600:
	case 1200:
	case 2400:
	case 4800:
	case 9600:
	case 19200:
	case 38400:
	case 57600:
	case 115200:
		break;
	default:
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Invalid serial port speed = %d", rs232->speed );

		rs232->speed = -1;

		return mx_status;
	}

	/* Check for allowed word sizes. */

	switch( rs232->word_size ) {
	case 7:
	case 8:
		break;
	default:
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		    "Invalid serial port word size = %d", rs232->word_size );

		rs232->word_size = -1;

		return mx_status;
	}

	/* Check for allowed parity. */

	switch( rs232->parity ) {
	case 'n':
	case 'N':
		rs232->parity = MXF_232_NO_PARITY;
		break;
	case 'e':
	case 'E':
		rs232->parity = MXF_232_EVEN_PARITY;
		break;
	case 'o':
	case 'O':
		rs232->parity = MXF_232_ODD_PARITY;
		break;
	case 'm':
	case 'M':
		rs232->parity = MXF_232_MARK_PARITY;
		break;
	case 's':
	case 'S':
		rs232->parity = MXF_232_SPACE_PARITY;
		break;
	default:
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Invalid serial port parity = '%c'", rs232->parity );

		rs232->parity = -1;

		return mx_status;
	}

	/* Check for allowed stop bits. */

	switch( rs232->stop_bits ) {
	case 1:
	case 2:
		break;
	default:
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		    "Invalid serial port stop bits = %d", rs232->stop_bits );

		rs232->stop_bits = -1;

		return mx_status;
	}

	/* Check for allowed flow control. */

	switch( rs232->flow_control ) {
	case 'n':
	case 'N':
		rs232->flow_control = MXF_232_NO_FLOW_CONTROL;
		break;

	case 'h':
	case 'H':
		rs232->flow_control = MXF_232_HARDWARE_FLOW_CONTROL;
		break;

	case 's':
	case 'S':
		rs232->flow_control = MXF_232_SOFTWARE_FLOW_CONTROL;
		break;

	case 'b':
	case 'B':
		rs232->flow_control = MXF_232_BOTH_FLOW_CONTROL;
		break;

	case 'x':
	case 'X':
		rs232->flow_control = MXF_232_XON_XOFF_FLOW_CONTROL;
		break;

	case 'r':
	case 'R':
		rs232->flow_control = MXF_232_RTS_CTS_FLOW_CONTROL;
		break;

	case 'd':
	case 'D':
		rs232->flow_control = MXF_232_DTR_DSR_FLOW_CONTROL;
		break;

	default:
		mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Invalid serial port flow control = '%s'",
			rs232->flow_control );

		rs232->flow_control = -1;

		return mx_status;
	}

	/* Save the value of the line terminators. */

	for ( i = 0, j = 0; i < MX_RS232_MAX_TERMINATORS; i++ ) {

		rs232->read_terminator_array[i] = '\0';

		shift = 8 * ( MX_RS232_MAX_TERMINATORS - i - 1 );

		terminator_value = (unsigned char)
				( ( rs232->read_terminators ) >> shift );

		if ( ( j > 0 ) || ( terminator_value != '\0' ) ) {

			rs232->read_terminator_array[j] = terminator_value;

			j++;
		}
	}

	rs232->num_read_terminator_chars = j;

	MX_DEBUG( 1,("%s: rs232->num_read_terminator_chars = %d",
			fname, rs232->num_read_terminator_chars));

#if 0
	MX_DEBUG(-2,("%s: read terminators = 0x%08lx",
			fname, rs232->read_terminators));

	for ( i = 0; i < 4; i++ ) {
		MX_DEBUG(-2,("%s: read_terminator_array[%d] = %x",
			fname, i, rs232->read_terminator_array[i] ));
	}
#endif

	for ( i = 0, j = 0; i < MX_RS232_MAX_TERMINATORS; i++ ) {

		rs232->write_terminator_array[i] = '\0';

		shift = 8 * ( MX_RS232_MAX_TERMINATORS - i - 1 );

		terminator_value = (unsigned char)
				( ( rs232->write_terminators ) >> shift );

		if ( ( j > 0 ) || ( terminator_value != '\0' ) ) {

			rs232->write_terminator_array[j] = terminator_value;

			j++;
		}
	}

	rs232->num_write_terminator_chars = j;

	MX_DEBUG( 1,("%s: rs232->num_write_terminator_chars = %d",
			fname, rs232->num_write_terminator_chars));

#if 0
	MX_DEBUG(-2,("%s: write terminators = 0x%08lx",
			fname, rs232->write_terminators));

	for ( i = 0; i < 4; i++ ) {
		MX_DEBUG(-2,("%s: write_terminator_array[%d] = %x",
			fname, i, rs232->write_terminator_array[i] ));
	}
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_rs232_get_pointers( MX_RECORD *rs232_record,
			MX_RS232 **rs232,
			MX_RS232_FUNCTION_LIST **function_list_ptr,
			const char *calling_fname )
{
	const char fname[] = "mx_rs232_get_pointers()";

	if ( rs232_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The rs232_record pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( rs232_record->mx_class != MXI_RS232 ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The record '%s' passed by '%s' is not a RS232 interface.  "
	"(superclass = %ld, class = %ld, type = %ld)",
			rs232_record->name, calling_fname,
			rs232_record->mx_superclass,
			rs232_record->mx_class,
			rs232_record->mx_type );
	}

	if ( rs232 != (MX_RS232 **) NULL ) {
		*rs232 = (MX_RS232 *) (rs232_record->record_class_struct);

		if ( *rs232 == (MX_RS232 *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RS232 pointer for record '%s' passed by '%s' is NULL.",
				rs232_record->name, calling_fname );
		}
	}

	if ( function_list_ptr != (MX_RS232_FUNCTION_LIST **) NULL ) {
		*function_list_ptr = (MX_RS232_FUNCTION_LIST *)
				(rs232_record->class_specific_function_list);

		if ( *function_list_ptr == (MX_RS232_FUNCTION_LIST *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_RS232_FUNCTION_LIST pointer for record '%s' passed by '%s' is NULL.",
				rs232_record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_rs232_getchar( MX_RECORD *record, char *c, int flags )
{
	const char fname[] = "mx_rs232_getchar()";

	MX_RS232 *rs232;
	MX_RS232_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_RS232 *, char *, int );
	mx_status_type status;

	status = mx_rs232_get_pointers( record, &rs232, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = ( fl_ptr->getchar );

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"close function pointer is NULL.");
	}

	/* Invoke the function. */

	status = (*fptr)( rs232, c, flags );

	return status;
}
		
MX_EXPORT mx_status_type
mx_rs232_putchar( MX_RECORD *record, char c, int flags )
{
	const char fname[] = "mx_rs232_putchar()";

	MX_RS232 *rs232;
	MX_RS232_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_RS232 *, char, int );
	mx_status_type status;

	status = mx_rs232_get_pointers( record, &rs232, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = ( fl_ptr->putchar );

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"close function pointer is NULL.");
	}

	/* Invoke the function. */

	status = (*fptr)( rs232, c, flags );

	return status;
}
		
MX_EXPORT mx_status_type
mx_rs232_read( MX_RECORD *record,
		char *buffer,
		size_t max_bytes_to_read,
		size_t *bytes_read,
		int flags )
{
	const char fname[] = "mx_rs232_read()";

	MX_RS232 *rs232;
	MX_RS232_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_RS232 *, char *, size_t, size_t * );
	size_t bytes_read_by_driver;
	mx_status_type status;

	status = mx_rs232_get_pointers( record, &rs232, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = ( fl_ptr->read );

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"close function pointer is NULL.");
	}

	/* Invoke the function. */

	status = (*fptr)( rs232, buffer,
			max_bytes_to_read, &bytes_read_by_driver );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( flags ) {
		MX_DEBUG(-2,("%s: received buffer = '%s'", fname, buffer));
	}

	if ( bytes_read != NULL ) {
		*bytes_read = bytes_read_by_driver;
	}

	return MX_SUCCESSFUL_RESULT;
}
		
MX_EXPORT mx_status_type
mx_rs232_write( MX_RECORD *record,
		char *buffer,
		size_t bytes_to_write,
		size_t *bytes_written,
		int flags )
{
	const char fname[] = "mx_rs232_write()";

	MX_RS232 *rs232;
	MX_RS232_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_RS232 *, char *, size_t, size_t * );
	size_t bytes_written_by_driver;
	mx_status_type status;

	status = mx_rs232_get_pointers( record, &rs232, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = ( fl_ptr->write );

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"close function pointer is NULL.");
	}

	if ( flags ) {
		MX_DEBUG(-2, ("%s: sending buffer = '%s'", fname, buffer));
	}

	/* Invoke the function. */

	status = (*fptr)( rs232, buffer,
				bytes_to_write, &bytes_written_by_driver );

	if ( bytes_written != NULL ) {
		*bytes_written = bytes_written_by_driver;
	}

	return status;
}

/* mx_rs232_getline() and mx_rs232_putline() are intended for reading
 * ASCII character strings.  Use mx_rs232_read() and mx_rs232_write()
 * for binary data.
 */

MX_EXPORT mx_status_type
mx_rs232_getline( MX_RECORD *record, char *buffer, size_t max_bytes_to_read,
			size_t *bytes_read, int flags )
{
	const char fname[] = "mx_rs232_getline()";

	MX_RS232 *rs232;
	MX_RS232_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_RS232 *, char *, size_t, size_t * );
	char *array;
	int i, terminators_seen, start_of_terminator;
	char c;
	mx_status_type status;

	status = mx_rs232_get_pointers( record, &rs232, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Does this driver have a getline function? */

	fptr = fl_ptr->getline;

	if ( fptr != NULL ) {

		/* If so, invoke it. */

		status = (*fptr)(rs232, buffer, max_bytes_to_read, bytes_read);

	} else {
		/* Otherwise, handle input a character at a time. */

		array = rs232->read_terminator_array;

		terminators_seen = 0;
		start_of_terminator = INT_MAX;

		for ( i = 0; i < max_bytes_to_read; i++ ) {
			status = mx_rs232_getchar(record, &c, MXF_232_WAIT);

			if ( status.code != MXE_SUCCESS ) {
	                        /* Make the buffer contents a valid C string
				 * before returning.
				 */

				buffer[i] = '\0';

				if ( status.code != MXE_NOT_READY ) {
					if ( flags ) {
						MX_DEBUG(-2,
						("%s failed.\nbuffer = '%s'",
							fname, buffer));

						MX_DEBUG(-2,
				("Failed with status = %d, c = 0x%x '%c'",
							status, c, c));
					}
					return status;
				}
			}
#if 0
			if ( flags ) {
				MX_DEBUG(-2,
				("%s: received c = 0x%x '%c'", fname, c, c));
			}
#endif

			buffer[i] = c;

			/* Check to see if we are in the middle of a 
			 * line terminator sequence.
			 */

			if ( c != array[ terminators_seen ] ) {

				terminators_seen = 0;
				start_of_terminator = INT_MAX;
			} else {
				if ( terminators_seen == 0 ) {
					start_of_terminator = i;
				}
				terminators_seen++;

				/* Only allow up to four terminator characters.
				 */

				if ( (terminators_seen
						>= MX_RS232_MAX_TERMINATORS)
				  || (array[ terminators_seen ] == '\0') ) {

					/* Since we have now seen the entire
					 * terminator sequence, we know that
					 * we are at the end of the line.
					 * Thus, we break out of the loop here.
					 */

					break;	/* Exit the for() loop. */
				}
			}
		}

		if ( start_of_terminator < i ) {
			buffer[ start_of_terminator ] = 0;
		} else {
			buffer[ i ] = 0;
		}

		if ( bytes_read != NULL ) {
			*bytes_read = strlen( buffer );
		}

		status = MX_SUCCESSFUL_RESULT;
	}

	if ( flags ) {
		MX_DEBUG(-2,("%s: received buffer = '%s'", fname, buffer));
	}

	return status;
}

MX_EXPORT mx_status_type
mx_rs232_putline( MX_RECORD *record,
		char *buffer,
		size_t *bytes_written,
		int flags )
{
	const char fname[] = "mx_rs232_putline()";

	MX_RS232 *rs232;
	MX_RS232_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_RS232 *, char *, size_t * );
	char *array;
	int i, length;
	char c;
	mx_status_type status;

	status = mx_rs232_get_pointers( record, &rs232, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( flags ) {
		MX_DEBUG(-2, ("%s: sending buffer = '%s'", fname, buffer));
	}

	/* Does this driver have a putline function? */

	fptr = fl_ptr->putline;

	if ( fptr != NULL ) {

		/* If so, invoke it. */

		status = (*fptr)( rs232, buffer, bytes_written );

	} else {
		/* Otherwise, handle output a character at a time. */

		array = rs232->write_terminator_array;

		length = strlen( buffer );

		for ( i = 0; i < length; i++ ) {
			c = buffer[i];

			status = mx_rs232_putchar(record, c, MXF_232_WAIT);

			if ( status.code != MXE_SUCCESS )
				return status;
		}

		/* Send the terminator characters. */

		for ( i = 0; i < MX_RS232_MAX_TERMINATORS; i++ ) {

			c = array[i];

			if ( c == '\0' ) {
				/* Have reached the end of the terminators,
				 * so exit the loop.
				 */

				break;
			}

			if ( flags ) {
				MX_DEBUG(-2,
				("%s: sending write terminator 0x%x",
					fname, c));
			}

			status = mx_rs232_putchar(record, c, MXF_232_WAIT);

			if ( status.code != MXE_SUCCESS )
				return status;
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_rs232_input_is_available( MX_RECORD *record, int *input_available )
{
	const char fname[] = "mx_rs232_input_is_available()";

	MX_RS232 *rs232;
	MX_RS232_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_RS232 *, int * );
	mx_status_type status;

	status = mx_rs232_get_pointers( record, &rs232, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = ( fl_ptr->input_is_available );

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"close function pointer is NULL.");
	}

	/* Invoke the function. */

	status = (*fptr)( rs232, input_available );

	return status;
}

MX_EXPORT mx_status_type
mx_rs232_discard_unread_input( MX_RECORD *record, int debug_flag )
{
	const char fname[] = "mx_rs232_discard_unread_input()";

	MX_RS232 *rs232;
	MX_RS232_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_RS232 *, int );
	mx_status_type status;

	status = mx_rs232_get_pointers( record, &rs232, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = ( fl_ptr->discard_unread_input );

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"close function pointer is NULL.");
	}

	/* Invoke the function. */

	status = (*fptr)( rs232, debug_flag );

	return status;
}

MX_EXPORT mx_status_type
mx_rs232_discard_unwritten_output( MX_RECORD *record, int debug_flag )
{
	const char fname[] = "mx_rs232_discard_unwritten_output()";

	MX_RS232 *rs232;
	MX_RS232_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_RS232 *, int );
	mx_status_type status;

	status = mx_rs232_get_pointers( record, &rs232, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = ( fl_ptr->discard_unwritten_output );

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"close function pointer is NULL.");
	}

	/* Invoke the function. */

	status = (*fptr)( rs232, debug_flag );

	return status;
}

