/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/* mmeasure.c -- Measure dark currents function. */

/* Currently this function expects that the scalers are gated on and off
 * by the timer called "timer1".  Probably, there should be something 
 * defined in the MX_SCALER structure that tells you which timer controls
 * which scaler, but currently there is not.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "motor.h"
#include "mdialog.h"

#define FREE_SCALER_ARRAYS \
	do { \
		if ( scaler_array != NULL ) { \
			free( scaler_array ); \
			scaler_array = NULL; \
		} \
		if ( scaler_value_array != NULL ) { \
			free( scaler_value_array ); \
			scaler_value_array = NULL; \
		} \
	} while(0)

int
motor_measure_fn( int argc, char *argv[] )
{
	static char usage[] = "Usage:  measure dark_currents\n";

	MX_RECORD *timer_record;
	MX_RECORD *current_record;
	MX_RECORD **scaler_array;
	long *scaler_value_array;
	long i, num_scalers, value;
	double measurement_time;
#if 0
	double seconds_left;
#endif
	long n, num_measurements;
	double new_dark_current;
	size_t length;
	int status, busy;
	mx_status_type mx_status;

	scaler_array = NULL;
	scaler_value_array = NULL;

	if ( argc <= 2 ) {
		fprintf( output, "%s\n", usage );
		return FAILURE;
	}

	length = strlen( argv[2] );

	if ( length <= 0 )
		length = 1;

	if ( (strncmp( argv[2], "dark_currents", length ) != 0)
	  && (strncmp( argv[2], "pedestals", length ) != 0) ) {
		fprintf( output, "\n" );
		fprintf( output,
"Only dark current measurements are currently supported by this command.\n\n");
		fprintf( output, "%s\n", usage );
		return FAILURE;
	}

	/* Verify that the "timer1" record exists, since we can't make
	 * a measurement without it.
	 */

	timer_record = mx_get_record( motor_record_list, "timer1" );

	if ( timer_record == (MX_RECORD *) NULL ) {
		fprintf(output,"The timer record 'timer1' does not exist!\n");
		fprintf(output,"We cannot make a measurement without it.\n");
		return FAILURE;
	}

	/* Get a list of all the scaler records. */

	/* First, how many scalers are there? */

	num_scalers = 0L;

	current_record = motor_record_list->next_record;

	while ( current_record != motor_record_list ) {
		if ( current_record->mx_class == MXC_SCALER ) {
			num_scalers++;
		}
		current_record = current_record->next_record;
	}

	/* Allocate an array to store the list of scalers. */

	scaler_array = (MX_RECORD **) malloc(num_scalers * sizeof(MX_RECORD *));

	if ( scaler_array == (MX_RECORD **) NULL ) {
		fprintf( output,
		"Ran out of memory trying to allocate an array of %ld scaler "
		"record pointers.\n", num_scalers );
		return FAILURE;
	}

	scaler_value_array = (long *) malloc( num_scalers * sizeof(long) );

	if ( scaler_value_array == (long *) NULL ) {
		fprintf( output,
"Ran out of memory trying to allocate an array of %ld scaler values.\n",
			num_scalers );
		FREE_SCALER_ARRAYS;
		return FAILURE;
	}

	/* Now fill the array with the MX_SCALER pointers. */

	i = 0;

	current_record = motor_record_list->next_record;

	while ( (current_record != motor_record_list) && (i < num_scalers) ) {

		if ( current_record->mx_class == MXC_SCALER ) {
			scaler_array[i] = current_record;

			i++;
		}
		current_record = current_record->next_record;
	}

	if ( i < num_scalers ) {
		fprintf( output,
"Warning: The number of scalers changed during startup of this command.\n" );
		fprintf( output, "num_scalers = %ld, i = %ld\n",num_scalers,i);
	}

#if 0
	fprintf( output, "Number of scalers = %ld\n", num_scalers );
#endif

	/* Prompt the user for the measurement parameters. */

	if ( argc > 3 ) {
		measurement_time = atof( argv[3] );
	} else {
		measurement_time = 0.0;

		status = motor_get_double( output,
			"Enter measurement time (seconds) -> ",
			TRUE, 1.0,
			&measurement_time, 0.0, 1.0e38 );

		if ( status != SUCCESS ) {
			FREE_SCALER_ARRAYS;
			return status;
		}
	}

	if ( argc > 4 ) {
		num_measurements = atol( argv[4] );
	} else {
		num_measurements = 0L;

		status = motor_get_long( output,
			"Enter number of measurements to average -> ",
			TRUE, 1,
			&num_measurements, 1L, 1000000L );

		if ( status != SUCCESS ) {
			FREE_SCALER_ARRAYS;
			return status;
		}
	}

	/***** Now, perform the measurements. *****/

	/* First, zero the scaler value array. */

	for ( i = 0; i < num_scalers; i++ ) {
		scaler_value_array[i] = 0L;
	}

#if 0
	/* Stop the timer if it is on. */

	mx_status = mx_timer_stop( timer_record, &seconds_left );

	if ( mx_status.code != MXE_SUCCESS ) {
		FREE_SCALER_ARRAYS;
		return FAILURE;
	}
#endif

	/* Loop through the number of measurements. */

	for ( n = 0; n < num_measurements; n++ ) {

		/* Clear all the scalers. */

		for ( i = 0; i < num_scalers; i++ ) {
			mx_status = mx_scaler_clear( scaler_array[i] );

			if ( mx_status.code != MXE_SUCCESS ) {
				FREE_SCALER_ARRAYS;
				return FAILURE;
			}
		}

		/* Turn the timer on and wait for it to finish. */

		mx_status = mx_timer_start( timer_record, measurement_time );

		if ( mx_status.code != MXE_SUCCESS ) {
			FREE_SCALER_ARRAYS;
			return FAILURE;
		}

		busy = TRUE;

		while ( busy ) {
			if ( mx_user_requested_interrupt() ) {
				fprintf( output,
				"Dark current measurement was interrupted.\n");
				FREE_SCALER_ARRAYS;
				return FAILURE;
			}

			mx_status = mx_timer_is_busy( timer_record, &busy );

			if ( mx_status.code != MXE_SUCCESS ) {
				FREE_SCALER_ARRAYS;
				return FAILURE;
			}

			mx_msleep(1);	/* Wait at least 1 millisecond. */
		}

		/* Read out the scalers. */

		for ( i = 0; i < num_scalers; i++ ) {
			mx_status = mx_scaler_read( scaler_array[i], &value );

			if ( mx_status.code != MXE_SUCCESS ) {
				FREE_SCALER_ARRAYS;
				return FAILURE;
			}

			scaler_value_array[i] += value;
		}
	}

	/* Now save the new dark_current values. */

	for ( i = 0; i < num_scalers; i++ ) {
		new_dark_current = (double) scaler_value_array[i]
			/ (measurement_time * (double) num_measurements);

		mx_status = mx_scaler_set_dark_current( scaler_array[i],
					new_dark_current );

		if ( mx_status.code != MXE_SUCCESS ) {
			FREE_SCALER_ARRAYS;
			return FAILURE;
		}

		fprintf( output,
			"Dark current for '%s' = %.*g counts per second.\n",
			scaler_array[i]->name,
			scaler_array[i]->precision,
			new_dark_current );
	}

	FREE_SCALER_ARRAYS;

	return SUCCESS;
}

