/*
 * Name:    ms_prscl.c
 *
 * Purpose: Functions used to process MX scaler record events.
 *
 *------------------------------------------------------------------------
 *
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_rec.h"
#include "mx_scalr.h"
#include "mx_sock.h"

#include "ms_serv.h"
#include "ms_prscl.h"

mx_status_type
mxsrv_setup_scaler_process_functions( MX_RECORD *record_list_head )
{
	const char fname[] = "mxsrv_setup_scaler_process_functions()";

	MX_RECORD *current_record;
	MX_RECORD_FIELD *record_field;
	MX_RECORD_FIELD *record_field_array;
	long i;

	MX_DEBUG(2,("%s invoked.", fname));

	/* The following is done just in case we were called with
	 * a record that isn't really the list head.
	 */

	record_list_head = record_list_head->list_head;

	/* Go through the record list looking for scaler records. */

	current_record = record_list_head->next_record;

	while ( current_record != record_list_head ) {

		if ( (current_record->mx_superclass == MXR_DEVICE)
		  && (current_record->mx_class == MXC_SCALER) ) {

			record_field_array
				= current_record->record_field_array;

			for ( i = 0;
				i < current_record->num_record_fields;
				i++ )
			{
				record_field = &record_field_array[i];

				switch( record_field->label_value ) {
				case MXLV_SCL_VALUE:
				case MXLV_SCL_CLEAR:
				case MXLV_SCL_OVERFLOW_SET:
				case MXLV_SCL_BUSY:
				case MXLV_SCL_STOP:
				case MXLV_SCL_MODE:
					record_field->process_function
					    = mxsrv_scaler_process_function;
					break;
				default:
					break;
				}
			}
		}
		current_record = current_record->next_record;
	}

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_scaler_process_function( void *record_ptr,
			void *record_field_ptr, int operation )
{
	const char fname[] = "mxsrv_scaler_process_function()";

	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	MX_SCALER *scaler;
	char operation_string[30];
	long label_value;
	long value;
	int overflow_set, busy, mode;
	mx_status_type status;

	record = (MX_RECORD *) record_ptr;
	record_field = (MX_RECORD_FIELD *) record_field_ptr;

	switch ( operation ) {
	case MX_PROCESS_GET:
		strcpy( operation_string, "(MX_PROCESS_GET)" );
		break;
	case MX_PROCESS_PUT:
		strcpy( operation_string, "(MX_PROCESS_PUT)" );
		break;
	default:
		sprintf( operation_string, "(Operation %d)", operation );
		break;
	}

	label_value = record_field->label_value;

	MX_DEBUG(1,("%s: '%s.%s' %s (%ld)",
		fname, record->name, record_field->name,
		operation_string, label_value));

	/*
	 * A label value < 0 tells us that we do not need to do any
	 * further MX record processing. 
	 */

	if ( label_value < 0 ) {
		return MX_SUCCESSFUL_RESULT;
	}

	scaler = (MX_SCALER *) (record->record_class_struct);

	status = MX_SUCCESSFUL_RESULT;

	switch( operation ) {
	case MX_PROCESS_GET:
		switch( label_value ) {
		case MXLV_SCL_VALUE:
			status = mx_scaler_read( record, &value );
			break;
		case MXLV_SCL_OVERFLOW_SET:
			status = mx_scaler_overflow_set(record, &overflow_set);
			break;
		case MXLV_SCL_BUSY:
			status = mx_scaler_is_busy( record, &busy );
			break;
		case MXLV_SCL_MODE:
			status = mx_scaler_get_mode( record, &mode );
			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_GET label value = %ld",
				fname, label_value));
			break;
		}

		/* Now invoke the value changed callback. */

		if ( status.code == MXE_SUCCESS ) {
			status = mxsrv_invoke_callback_list(
					MXSF_VALUE_CHANGED_CALLBACK,
					record_field );
		}
		break;
	case MX_PROCESS_PUT:
		switch( label_value ) {
		case MXLV_SCL_VALUE:
			status = mx_scaler_start( record, scaler->value );
			break;
		case MXLV_SCL_CLEAR:
			status = mx_scaler_clear( record );
			break;
		case MXLV_SCL_STOP:
			status = mx_scaler_stop( record, &value );
			break;
		case MXLV_SCL_MODE:
			status = mx_scaler_set_mode( record, scaler->mode );
			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_PUT label value = %ld",
				fname, label_value));
			break;
		}

		/* Now invoke the value changed callback. */

		if ( status.code == MXE_SUCCESS ) {
			status = mxsrv_invoke_callback_list(
					MXSF_VALUE_CHANGED_CALLBACK,
					record_field );
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown operation code = %ld", operation );
		break;
	}

	return status;
}

